"""
Helper methods for class for working with Pilatus images, for instance for
identifying the regions to be masked.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class _Detector:
    module_size_fast: int
    module_size_slow: int
    gap_fast: int
    gap_slow: int
    module_configs_fast: tuple[int, ...]

    @property
    def all_widths(self) -> list[int]:
        return [
            self.module_size_fast * j + self.gap_fast * (j - 1)
            for j in self.module_configs_fast
        ]


_DetectorDatabase = {
    "Pilatus": _Detector(
        module_size_fast=487,
        module_size_slow=195,
        gap_fast=7,
        gap_slow=17,
        module_configs_fast=(1, 3, 5),
    ),
    "Eiger X": _Detector(
        module_size_fast=1030,
        module_size_slow=514,
        gap_fast=10,
        gap_slow=37,
        module_configs_fast=(1, 2, 3, 4),
    ),
    "Eiger 2X": _Detector(
        module_size_fast=1028,
        module_size_slow=512,
        gap_fast=12,
        gap_slow=38,
        module_configs_fast=(1, 2, 3, 4),
    ),
}


def pilatus_6M_mask():
    """Hard coded mask regions for a Pilatus 6M instrument."""
    # FIX me, the parameters are listed here as f0, f1, s0, s1 but the prototype specifies f0, s0, f1, s1
    return [
        [488, 494, 1, 2527],
        [982, 988, 1, 2527],
        [1476, 1482, 1, 2527],
        [1970, 1976, 1, 2527],
        [1, 2463, 196, 212],
        [1, 2463, 408, 424],
        [1, 2463, 620, 636],
        [1, 2463, 832, 848],
        [1, 2463, 1044, 1060],
        [1, 2463, 1256, 1272],
        [1, 2463, 1468, 1484],
        [1, 2463, 1680, 1696],
        [1, 2463, 1892, 1908],
        [1, 2463, 2104, 2120],
        [1, 2463, 2316, 2332],
    ]


def pilatus_2M_mask():
    """Hard coded mask regions for a Pilatus 2M detector."""

    return [
        [488, 494, 1, 1679],
        [982, 988, 1, 1679],
        [1, 1475, 196, 212],
        [1, 1475, 408, 424],
        [1, 1475, 620, 636],
        [1, 1475, 832, 848],
        [1, 1475, 1044, 1060],
        [1, 1475, 1256, 1272],
        [1, 1475, 1468, 1484],
    ]


def pilatus_300K_mask():
    """Hard coded mask regions for a Pilatus 300K instrument."""

    return [[1, 487, 196, 212], [1, 487, 408, 424]]


def _get_pad_module_gap(xdetector, size=None):
    assert len(xdetector) == 1
    assert xdetector[0].get_type() == "SENSOR_PAD"

    if size is None:
        size = xdetector[0].get_image_size()

    for detector in _DetectorDatabase.values():
        if size[0] in detector.all_widths:
            return detector


def sensor_active_areas(xdetector):
    """Return the sensitive areas on the detector for pixel array detectors
    yes, does include hard coded magic numbers; returns [(x0, y0 x1, x1)]"""

    size = xdetector[0].get_image_size()
    detector = _get_pad_module_gap(xdetector)

    # iterate over these to produce a list of the TLC, BRC of the modules
    # as a list
    n_fast, remainder = divmod(size[0], detector.module_size_fast)
    assert (n_fast - 1) * detector.gap_fast == remainder

    n_slow, remainder = divmod(size[1], detector.module_size_slow)
    assert (n_slow - 1) * detector.gap_slow == remainder

    # Specify the dead areas between the modules, i.e. the rows and columns
    # where there are no active pixels

    panels = []

    for i_fast in range(n_fast):
        for i_slow in range(n_slow):
            panels.append(
                (
                    i_fast * detector.module_size_fast,
                    i_slow * detector.module_size_slow,
                    (i_fast + 1) * detector.module_size_fast,
                    (i_slow + 1) * detector.module_size_slow,
                )
            )

    return panels


def determine_pilatus_mask(xdetector):
    """Return an appropriate pixel mask for a Pilatus detector."""

    size = xdetector[0].get_image_size()

    # Hardcoded module size and gap size
    detector = _DetectorDatabase["Pilatus"]

    # Edge dead areas not included, only gaps between modules matter
    n_fast, remainder = divmod(size[0], detector.module_size_fast)
    assert (n_fast - 1) * detector.gap_fast == remainder

    n_slow, remainder = divmod(size[1], detector.module_size_slow)
    assert (n_slow - 1) * detector.gap_slow == remainder

    # Specify the dead areas between the modules, i.e. the rows and columns
    # where there are no active pixels
    mask = []
    for i_fast in range(n_fast - 1):
        mask.append(
            [
                (i_fast + 1) * detector.module_size_fast
                + i_fast * detector.gap_fast
                + 1,
                (i_fast + 1) * detector.module_size_fast
                + i_fast * detector.gap_fast
                + detector.gap_fast,
                1,
                size[1],
            ]
        )
    for i_slow in range(n_slow - 1):
        mask.append(
            [
                1,
                size[0],
                (i_slow + 1) * detector.module_size_slow
                + i_slow * detector.gap_slow
                + 1,
                (i_slow + 1) * detector.module_size_slow
                + i_slow * detector.gap_slow
                + detector.gap_slow,
            ]
        )

    return mask


def determine_eiger_mask(xdetector):
    """Return an appropriate pixel mask for an Eiger detector."""

    size = xdetector[0].get_image_size()

    detector = _get_pad_module_gap(xdetector)

    if not detector:
        size = tuple(reversed(xdetector[0].get_image_size()))
        detector = _get_pad_module_gap(xdetector, size=size)

    # Edge dead areas not included, only gaps between modules matter
    n_fast, remainder = divmod(size[0], detector.module_size_fast)
    assert (n_fast - 1) * detector.gap_fast == remainder

    n_slow, remainder = divmod(size[1], detector.module_size_slow)
    assert (n_slow - 1) * detector.gap_slow == remainder

    # Specify the dead areas between the modules, i.e. the rows and columns
    # where there are no active pixels
    mask = []
    for i_fast in range(n_fast - 1):
        mask.append(
            [
                (i_fast + 1) * detector.module_size_fast
                + i_fast * detector.gap_fast
                + 1,
                (i_fast + 1) * detector.module_size_fast
                + i_fast * detector.gap_fast
                + detector.gap_fast,
                1,
                size[1],
            ]
        )
    for i_slow in range(n_slow - 1):
        mask.append(
            [
                1,
                size[0],
                (i_slow + 1) * detector.module_size_slow
                + i_slow * detector.gap_slow
                + 1,
                (i_slow + 1) * detector.module_size_slow
                + i_slow * detector.gap_slow
                + detector.gap_slow,
            ]
        )

    return mask


def get_vendortype(xdetector):
    array = xdetector[0].get_image_size()
    if array == (2463, 2527):
        return "Pilatus-6M"
    elif array == (2463, 195):
        # special treatment of Pilatus-12M, treat as -6M for the viewer
        return "Pilatus-6M"
    elif array == (1475, 1679):
        return "Pilatus-2M"
    elif array == (487, 619):
        return "Pilatus-300K"
    return "Undetermined Pilatus size"


def get_vendortype_eiger(xdetector):
    array = xdetector[0].get_image_size()
    # print array,
    if array == (4150, 4371):
        return "Eiger-16M"
    elif array == (3110, 3269):
        return "Eiger-9M"
    elif array == (2070, 2167):
        return "Eiger-4M"
    elif array == (1030, 1065):
        return "Eiger-1M"
    elif array == (1030, 514):
        return "Eiger-500K"
    elif array == (4148, 4362):
        return "Eiger2-16M"
    elif array == (3108, 3262):
        return "Eiger2-9M"
    elif array == (2068, 2162):
        return "Eiger2-4M"
    elif array == (1028, 1062):
        return "Eiger2-1M"
    elif array == (1028, 512):
        return "Eiger2-500K"
    return "Undetermined EigerX size"



from __future__ import annotations

import pycbf

from dxtbx.model.detector import DetectorFactory

try:
    from dxtbx.format.image import cbf_read_buffer
except ImportError:
    # If we didn't compile cbf_read_buffer, use the pycbf version
    def cbf_read_buffer(handle, contents, flags):
        """Use cbf_read_buffer via pycbf"""
        assert hasattr(pycbf.cbf_handle_struct, "read_buffer")
        return handle.read_buffer(contents, flags)


def test_cbf_buffer(dials_data):
    data_dir = dials_data("misc_regression")
    filename = str(data_dir / "dials-190_01_00001.cbf")
    with open(filename, "rb") as f:
        contents = f.read()

    handle = pycbf.cbf_handle_struct()
    cbf_read_buffer(handle, contents, pycbf.MSG_DIGEST)
    det = DetectorFactory.imgCIF_H(handle, "unknown")
    assert det



"""Format class for miniCBF files from an ELDICO ED-1 electron diffractometer,
which uses a DECTRIS QUADRO detector (EIGER technology)"""

from __future__ import annotations

from dxtbx.format.FormatCBFMiniEiger import FormatCBFMiniEiger
from dxtbx.model import SimplePxMmStrategy
from dxtbx.model.beam import Probe


class FormatCBFMiniEigerQuadroED1(FormatCBFMiniEiger):
    @staticmethod
    def understand(image_file):
        header = FormatCBFMiniEiger.get_cbf_header(image_file)

        for record in header.split("\n"):
            if record.startswith("# Detector") and not record.startswith("# Detector_"):
                record = record.lower()
                if not ("quadro" in record or "eiger" in record):
                    return False

            # ED-1 has fixed energy of 160 keV
            if "# wavelength" in record:
                try:
                    wl = float(record.split()[-2])
                except ValueError:
                    return False
                if round(wl, 3) != 0.029:
                    return False

        # If we got this far, check also the mime header to ensure a 512*512 pixel image
        mime_header = ""
        in_binary_format_section = False
        with FormatCBFMiniEiger.open_file(image_file, "rb") as fh:
            for record in fh:
                record = record.decode()
                if "--CIF-BINARY-FORMAT-SECTION--" in record:
                    in_binary_format_section = True
                elif in_binary_format_section and record[0] == "X":
                    mime_header += record
                if in_binary_format_section and len(record.strip()) == 0:
                    # http://sourceforge.net/apps/trac/cbflib/wiki/ARRAY_DATA%20Category
                    #    In an imgCIF file, the encoded binary data begins after
                    #    the empty line terminating the header.
                    break

        for record in mime_header.split("\n"):
            if "Dimension" in record:
                if int(record.split()[-1]) != 512:
                    return False

        return True

    def _goniometer(self):
        """Vertical axis"""

        return self._goniometer_factory.known_axis((0, 1, 0))

    def _beam(self):
        """Ensure an unpolarised beam"""

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])
        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )

    def _detector(self):
        distance = float(self._cif_header_dictionary["Detector_distance"].split()[0])

        beam_xy = (
            self._cif_header_dictionary["Beam_xy"]
            .replace("(", "")
            .replace(")", "")
            .replace(",", "")
            .split()[:2]
        )

        beam_x, beam_y = map(float, beam_xy)

        pixel_xy = (
            self._cif_header_dictionary["Pixel_size"]
            .replace("m", "")
            .replace("x", "")
            .split()
        )

        pixel_x, pixel_y = map(float, pixel_xy)

        if "Silicon" in self._cif_header_dictionary:
            thickness = (
                float(self._cif_header_dictionary["Silicon"].split()[2]) * 1000.0
            )
            material = "Si"
        else:
            thickness = 0.450
            material = "Si"

        # These are always both 512
        nx = int(self._cif_header_dictionary["X-Binary-Size-Fastest-Dimension"])
        ny = int(self._cif_header_dictionary["X-Binary-Size-Second-Dimension"])

        if "Count_cutoff" in self._cif_header_dictionary:
            overload = int(self._cif_header_dictionary["Count_cutoff"].split()[0])
        else:
            overload = 1048576
        if overload == 0:
            # TODO To be checked
            overload = 1048576

        minimum_trusted_value = 0

        try:
            identifier = self._cif_header_dictionary["Detector"].encode()
        except KeyError:
            identifier = "Unknown Eiger"

        detector = self._detector_factory.simple(
            sensor="PAD",
            distance=distance * 1000.0,
            beam_centre=(beam_x * pixel_x * 1000.0, beam_y * pixel_y * 1000.0),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(1000 * pixel_x, 1000 * pixel_y),
            image_size=(nx, ny),
            trusted_range=(minimum_trusted_value, overload),
            mask=[],
        )

        # Here we set specifics: parallax correction and
        # QE correction are effectively disabled by setting the simple
        # pixel-to-millimetre strategy and a very high mu value. Gain is not
        # precisely known. We think it might be around 1.7, but here will leave
        # as 1.0 and expect the user to override at import.
        for panel in detector:
            panel.set_thickness(thickness)
            panel.set_material(material)
            panel.set_identifier(identifier)
            panel.set_px_mm_strategy(SimplePxMmStrategy())
            panel.set_mu(1e10)

        return detector



from __future__ import annotations

import pytest

from dxtbx.format.FormatCBFMini import FormatCBFMini
from dxtbx.model.compare import sequence_diff
from dxtbx.model.experiment_list import (
    BeamComparison,
    DetectorComparison,
    ExperimentListFactory,
    GoniometerComparison,
)


@pytest.mark.parametrize(
    "image_file",
    [
        ("ALS_831-q315r_lyso_001.img.bz2"),
        ("DLS_I02-X4_wide_M1S4_1_0001.cbf.bz2"),
    ],
)
def test_cbf_writer(image_file, dials_data, tmp_path):
    data_dir = dials_data("image_examples")
    filename = str(data_dir / image_file)
    imageset = ExperimentListFactory.from_filenames([filename])[0].imageset
    output_file = tmp_path / "image_0001.cbf"

    FormatCBFMini.as_file(
        imageset.get_detector(),
        imageset.get_beam(),
        imageset.get_goniometer(),
        imageset.get_scan(),
        imageset.get_raw_data(0)[0],
        output_file,
    )
    assert output_file.is_file()
    assert imageset.get_format_class()

    imageset2 = ExperimentListFactory.from_filenames([output_file])[0].imageset

    d_u_o = pytest.importorskip("dials.util.options")
    tolerance = d_u_o.tolerance_phil_scope.extract().tolerance

    print(sequence_diff(imageset, imageset2, tolerance=tolerance))

    assert BeamComparison()(imageset.get_beam(), imageset2.get_beam())

    # FormatCBFMini.as_file does not account for detectors where the lower value
    # of trusted_range is not 0. It is not clear if that is even possible to
    # express in the miniCBF header. So force the original detector
    # trusted_range to start from 0 for the purposes of this comparison
    max_trusted_value = imageset.get_detector()[0].get_trusted_range()[1]
    imageset.get_detector()[0].set_trusted_range((0, max_trusted_value))
    assert DetectorComparison(origin_tolerance=tolerance.detector.origin)(
        imageset.get_detector(), imageset2.get_detector()
    )
    assert GoniometerComparison()(imageset.get_goniometer(), imageset2.get_goniometer())
    s1 = imageset.get_scan()
    s2 = imageset.get_scan()
    assert s1.get_exposure_times() == s2.get_exposure_times()
    assert s1.get_oscillation() == s2.get_oscillation()
    assert s1.get_image_range() == s2.get_image_range()
    assert imageset.get_raw_data(0) == imageset2.get_raw_data(0)



from __future__ import annotations

import os

import dials_data.download

df = dials_data.download.DataFetcher()
data_dir = df("image_examples")
image_examples = []
for f in data_dir.iterdir():
    ext = os.path.splitext(f)[1]
    if ext == ".h5" and "_master." not in str(f):
        continue
    if ext in (".expt", ".json"):
        continue
    image_examples.append(str(f))



"""
Helpers for the detector class... this time enumerating all of the common
detector types, hashed by the sensor type, image dimensions and pixel
dimensions.
"""

from __future__ import annotations

import os
import sys

import dxtbx
from dxtbx.model.detector import DetectorFactory
from dxtbx.model.detector_helpers import detector_helper_sensors


class _detector_helpers_types:
    """A singleton class to help with identifying specific detectors used for
    macromolecular crystallography."""

    def __init__(self):
        detector_lib = os.path.join(
            os.path.split(dxtbx.__file__)[0], "data", "detectors.lib"
        )

        if not os.path.exists(detector_lib):
            raise RuntimeError("detector library not found")

        self._detectors = {}

        with open(detector_lib, encoding="ascii") as fh:
            for record in fh:
                if record.startswith(("Sensor", "-----")):
                    continue

                text = record.split("#")[0].strip()

                if not text:
                    continue

                tokens = text.split()

                assert len(tokens) == 6

                sensor = DetectorFactory.sensor(tokens[0])
                fast, slow, df, ds = map(int, tokens[1:5])

                self._detectors[(sensor, fast, slow, df, ds)] = tokens[5]

    def get(self, sensor, fast, slow, df, ds):
        """Look up a name for a detector with this sensor type (listed in
        detector_helpers) these image dimensions in the fast and slow
        directions, these corresponding pixel sizes, in microns (integers).
        If the sensor is unknown, all sensor types will be tested - be warned
        if there are duplicates."""

        sensor = DetectorFactory.sensor(sensor)

        if sensor == detector_helper_sensors.SENSOR_UNKNOWN:
            for s in detector_helper_sensors.all():
                try:
                    return self.get(s, fast, slow, df, ds)
                except ValueError:
                    pass

            raise ValueError(
                "detector %s %d %d %d %d unknown" % (sensor, fast, slow, df, ds)
            )

        if (sensor, fast, slow, df, ds) in self._detectors:
            return self._detectors[(sensor, fast, slow, df, ds)]

        # OK allow for small variations in the recorded pixel dimensions

        for ddf in -2, -1, 1, 2:
            for dds in -2, -1, 1, 2:
                if (sensor, fast, slow, df + ddf, ds + dds) in self._detectors:
                    return self._detectors[(sensor, fast, slow, df + ddf, ds + dds)]

        raise ValueError(
            "detector %s %d %d %d %d unknown" % (sensor, fast, slow, df, ds)
        )


detector_helpers_types = _detector_helpers_types()

if __name__ == "__main__":
    sensor = sys.argv[1]
    fast, slow, df, ds = map(int, sys.argv[2:6])

    print(detector_helpers_types.get(sensor, fast, slow, df, ds))



"""
Helpers for the scan class, which are things for handling e.g. filenames,
templates and so on.
"""

from __future__ import annotations

import math
import os
import re

# These are reversed patterns...
patterns = [
    r"([0-9]{2,12})\.(.*)",
    r"(.*)\.([0-9]{2,12})_(.*)",
    r"(.*)\.([0-9]{2,12})(.*)",
    r"(.*)\.([0-9]{1,12})([^0]*)_(.*)",
    r"(.*)\.([0-9]{1})(.*)",
]

joiners = [".", "_", "", "_", ""]

compiled_patterns = [re.compile(pattern) for pattern in patterns]


def template_regex(filename):
    """Try a bunch of templates to work out the most sensible. N.B. assumes
    that the image index will be the last digits found in the file name."""

    rfilename = filename[::-1]

    for j, cp in enumerate(compiled_patterns):
        match = cp.match(rfilename)
        if not match:
            continue
        groups = match.groups()

        if len(groups) == 3:
            exten = "." + groups[0][::-1]
            digits = groups[1][::-1]
            prefix = groups[2][::-1] + joiners[j]
        elif len(groups) == 4:
            exten = "." + groups[0][::-1]
            digits = groups[1][::-1]
            prefix = groups[3][::-1] + joiners[j] + groups[2][::-1]
        else:
            exten = ""
            digits = groups[0][::-1]
            prefix = groups[1][::-1] + joiners[j]

        template = prefix + ("#" * len(digits)) + exten
        return template, int(digits)

    # What is supposed to happen otherwise?
    raise ValueError("Could not determine filename template")


def _image2template(filename):
    return template_regex(filename)[0]


def _image2image(filename):
    return template_regex(filename)[1]


def _image2template_directory(filename):
    """Separate out the template and directory from an image name."""

    directory = os.path.dirname(filename)

    if not directory:
        # then it should be the current working directory
        directory = os.getcwd()

    image = os.path.split(filename)[-1]
    template = _image2template(image)

    return template, directory


def find_matching_images(template, directory):
    """Find images which match the input template in the directory
    provided."""

    files = os.listdir(directory)

    # to turn the template to a regular expression want to replace
    # however many #'s with EXACTLY the same number of [0-9] tokens,
    # e.g. ### -> ([0-9]{3})

    # change 30/may/2008 - now escape the template in this search to cope with
    # file templates with special characters in them, such as "+" -
    # fix to a problem reported by Joel B.

    length = template.count("#")
    regexp_text = re.escape(template).replace("\\#" * length, "([0-9]{%d})" % length)
    regexp = re.compile(regexp_text)

    images = []

    for f in files:
        match = regexp.match(f)

        if match:
            images.append(int(match.group(1)))

    images.sort()

    return images


def template_directory_number2image(template, directory, number):
    """Construct the full path to an image from the template, directory
    and image number."""

    # construct the full image name
    return os.path.join(directory, template_number2image(template, number))


def template_number2image(template, number):
    """Construct the an image from the template and image number."""

    length = template.count("#")

    # check that the number will fit in the template

    if (math.pow(10, length) - 1) < number:
        raise RuntimeError("number too big for template")

    format = "%%0%dd" % length

    return template.replace("#" * length, format % number)


class scan_helper_image_files:
    """A helper class which handles things like image names, making templates,
    finding matching images and so on. Currently this just provides aliases
    to existing functions elsewhere, but ultimately it would be good if they
    were all encapsulated herein."""

    @staticmethod
    def image_to_template(filename):
        """From an image name, return a file template which should match."""
        return _image2template(filename)

    @staticmethod
    def image_to_index(filename):
        """From an image name, determine the index within the scan for this
        image, complementary to the image_to_template method above."""
        return _image2image(filename)

    @staticmethod
    def image_to_template_directory(filename):
        """From a full path to an image, return the filename template and
        directory."""
        return _image2template_directory(filename)

    @staticmethod
    def template_directory_to_indices(template, directory):
        """For a given template and directory, return a list of image indices
        which match. Also complementary with image_to_template_directory."""
        return find_matching_images(template, directory)

    @staticmethod
    def template_directory_index_to_image(template, directory, index):
        """Construct the full image name from the template, directory and
        file index."""
        return template_directory_number2image(template, directory, index)

    @staticmethod
    def template_index_to_image(template, index):
        """Construct the image file name from the template and file index."""
        return template_number2image(template, index)



from __future__ import annotations

from cctbx import crystal_orientation
from cctbx.crystal import symmetry
from cctbx.uctbx import unit_cell
from libtbx.test_utils import approx_equal
from scitbx.matrix import sqr

from dxtbx.model import Crystal
from dxtbx.model.experiment_list import ExperimentListFactory

"""
This test exercises and demos several pieces of cctbx/dxtbx/nanoBragg
functionality:

1) Convert the known orientation matrix used to generate a nanoBragg image to
a dxtbx crystal object
2) Convert a dxtbx crystal object to a cctbx crystal_orientation object
3) Verify that these conversions are equivalent
4) Rotate an indexing solution onto the known orientation, accounting for
the indexing happening using an alternate symmetry operation
5) Verify the mis-orientation extent between the known and calculated
orientations

"""


GFC = symmetry(
    unit_cell=unit_cell((5.876, 7.299, 29.124, 90.00, 95.79, 90.00)),
    space_group_symbol="C 1 2/c 1",
)
CB_OP_C_P = GFC.change_of_basis_op_to_primitive_setting()  # from C to P

permute = sqr((0, 0, 1, 0, 1, 0, -1, 0, 0))

json_is = """{
  "__id__": "ExperimentList",
  "experiment": [
    {
      "__id__": "Experiment",
      "identifier": "",
      "beam": 0,
      "detector": 0,
      "crystal": 0,
      "imageset": 0
    }
  ],
  "imageset": [
    {
      "__id__": "ImageSet",
      "images": [
        "step5_000009.img.gz"
      ],
      "mask": "",
      "gain": "",
      "pedestal": "",
      "dx": "",
      "dy": "",
      "params": {}
    }
  ],
  "beam": [
    {
      "direction": [
        -0.0,
        -0.0,
        1.0
      ],
      "transmission": 1.0,
      "polarization_normal": [
        0.0,
        1.0,
        0.0
      ],
      "divergence": 0.0,
      "polarization_fraction": 0.999,
      "flux": 0.0,
      "sigma_divergence": 0.0,
      "wavelength": 1.30432
    }
  ],
  "detector": [
    {
      "hierarchy": {
        "origin": [
          0.0,
          0.0,
          0.0
        ],
        "fast_axis": [
          1.0,
          0.0,
          0.0
        ],
        "name": "",
        "raw_image_offset": [
          0,
          0
        ],
        "slow_axis": [
          0.0,
          1.0,
          0.0
        ],
        "material": "",
        "mask": [],
        "thickness": 0.0,
        "mu": 0.0,
        "gain": 1.0,
        "trusted_range": [
          0.0,
          0.0
        ],
        "image_size": [
          0,
          0
        ],
        "px_mm_strategy": {
          "type": "SimplePxMmStrategy"
        },
        "identifier": "",
        "type": "",
        "children": [
          {
            "panel": 0
          }
        ],
        "pixel_size": [
          0.0,
          0.0
        ]
      },
      "panels": [
        {
          "origin": [
            -97.24,
            97.13000000000001,
            -50.0
          ],
          "fast_axis": [
            1.0,
            0.0,
            0.0
          ],
          "name": "Panel",
          "raw_image_offset": [
            0,
            0
          ],
          "slow_axis": [
            0.0,
            -1.0,
            0.0
          ],
          "material": "",
          "mask": [],
          "thickness": 0.0,
          "mu": 0.0,
          "gain": 1.0,
          "trusted_range": [
            -9.0,
            65525.0
          ],
          "image_size": [
            1765,
            1765
          ],
          "px_mm_strategy": {
            "type": "SimplePxMmStrategy"
          },
          "identifier": "",
          "type": "SENSOR_CCD",
          "pixel_size": [
            0.11,
            0.11
          ]
        }
      ]
    }
  ],
  "goniometer": [],
  "scan": [],
  "crystal": [
    {
      "__id__": "crystal",
      "real_space_a": [
        -1.6148276313609897,
        -5.437716408522997,
        -1.4401497447540712
      ],
      "real_space_b": [
        -0.3944198624524956,
        1.9800095805737359,
        -7.033859173305951
      ],
      "real_space_c": [
        28.45936960940352,
        -4.754762088402217,
        -2.934294620730966
      ],
      "space_group_hall_symbol": "-C 2yc",
      "ML_half_mosaicity_deg": 0.3692704083777256,
      "ML_domain_size_ang": 416.04559892901716
    }
  ]
}"""


def test_compare_example():
    experiments = ExperimentListFactory.from_json(json_is, check_format=False)

    for experiment in experiments:
        header = {
            "DIM": "2",
            "DENZO_X_BEAM": "97.185",
            "RANK": "0",
            "PREFIX": "step5_000009",
            "BEAM_CENTER_Y": "97.13",
            "BEAM_CENTER_X": "97.13",
            "WAVELENGTH": "1.30432",
            "OSC_START": "0",
            "ADC_OFFSET": "10",
            "BYTE_ORDER": "little_endian",
            "DIRECT_SPACE_ABC": "2.7790989649304656,3.721227037283121,0.616256870237976,-4.231398741367156,1.730877297917864,1.0247061633019547,2.9848502901631253,-4.645083818143041,28.595825588147285",
            "OSC_RANGE": "0",
            "DIALS_ORIGIN": "-97.185,97.185,-50",
            "MOSFLM_CENTER_X": "97.13",
            "MOSFLM_CENTER_Y": "97.13",
            "CLOSE_DISTANCE": "50",
            "BEAMLINE": "fake",
            "TWOTHETA": "0",
            "ADXV_CENTER_Y": "96.965",
            "ADXV_CENTER_X": "97.185",
            "HEADER_BYTES": "1024",
            "DETECTOR_SN": "000",
            "DISTANCE": "50",
            "PHI": "0",
            "SIZE1": "1765",
            "SIZE2": "1765",
            "XDS_ORGX": "884",
            "XDS_ORGY": "884",
            "DENZO_Y_BEAM": "97.185",
            "TIME": "1",
            "TYPE": "unsigned_short",
            "PIXEL_SIZE": "0.11",
        }
        # the header of the simulated image, containing ground truth orientation

        rsabc = (
            sqr([float(v) for v in header["DIRECT_SPACE_ABC"].split(",")])
            * permute.inverse()
        )
        rsa = rsabc[0:3]
        rsb = rsabc[3:6]
        rsc = rsabc[6:9]
        header_cryst = Crystal(rsa, rsb, rsc, "P1").change_basis(CB_OP_C_P.inverse())
        header_cryst.set_space_group(experiment.crystal.get_space_group())
        print("Header crystal")
        print(header_cryst)

        expt_crystal = experiment.crystal
        print("Integrated crystal")
        print(expt_crystal)

        header_ori = crystal_orientation.crystal_orientation(
            header_cryst.get_A(), crystal_orientation.basis_type.reciprocal
        )
        expt_ori = crystal_orientation.crystal_orientation(
            expt_crystal.get_A(), crystal_orientation.basis_type.reciprocal
        )

        print("Converted to cctbx")
        header_ori.show()
        expt_ori.show()

        # assert the equivalence between dxtbx crystal object and the cctbx crystal_orientation object
        assert approx_equal(header_cryst.get_U(), header_ori.get_U_as_sqr())
        assert approx_equal(header_cryst.get_A(), header_ori.reciprocal_matrix())
        assert approx_equal(
            header_cryst.get_B(),
            sqr(header_ori.unit_cell().fractionalization_matrix()).transpose(),
        )

        assert approx_equal(expt_crystal.get_U(), expt_ori.get_U_as_sqr())
        assert approx_equal(expt_crystal.get_A(), expt_ori.reciprocal_matrix())
        assert approx_equal(
            expt_crystal.get_B(),
            sqr(expt_ori.unit_cell().fractionalization_matrix()).transpose(),
        )

        cb_op_align = sqr(expt_ori.best_similarity_transformation(header_ori, 50, 1))

        print("XYZ angles", cb_op_align.r3_rotation_matrix_as_x_y_z_angles(deg=True))
        aligned_ori = expt_ori.change_basis(cb_op_align)

        U_integrated = aligned_ori.get_U_as_sqr()
        U_ground_truth = header_ori.get_U_as_sqr()

        missetting_rot = U_integrated * U_ground_truth.inverse()
        print("determinant", missetting_rot.determinant())
        assert approx_equal(missetting_rot.determinant(), 1.0)
        assert missetting_rot.is_r3_rotation_matrix()

        (
            angle,
            axis,
        ) = missetting_rot.r3_rotation_matrix_as_unit_quaternion().unit_quaternion_as_axis_and_angle(
            deg=True
        )
        print("Angular offset is %13.10f deg." % (angle))

        assert approx_equal(angle, 0.2609472065)



"""
Base implementation of CBF formats - which is just really a place holder
which will tell you whether something is a CBF file (or no.)
"""

from __future__ import annotations

import sys

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format


class FormatCBF(Format):
    """An image reading class for CBF format images i.e. those from Dectris
    amongst others. This is just a first base class which will be used to
    determine whether this is really a CBF file."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        with FormatCBF.open_file(image_file, "rb") as fh:
            return b"###CBF" == fh.read(6)

    @staticmethod
    def get_cbf_header(image_file):
        """Obtain the text section of the header, which is assumed to be
        everything before --CIF-BINARY-FORMAT-SECTION-- - N.B. for reasons
        of simplicity will read the file in 4k chunks."""

        marker = b"--CIF-BINARY-FORMAT-SECTION--"

        with FormatCBF.open_file(image_file, "rb") as fin:
            header = fin.read(4096)
            marker_index = header.find(marker)
            marker_search_start = len(header) - len(marker) + 1
            while marker_index <= 0:
                add = fin.read(4096)
                if not add:
                    # If the marker is not contained in the file then we return the
                    # entire file.
                    marker_index = None
                    break
                header += add
                marker_index = header.find(marker, marker_search_start)
                marker_search_start += 4096
            header = header[:marker_index]
            return header.decode("ascii", "ignore")

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)

        super().__init__(str(image_file), **kwargs)

    @staticmethod
    def _parse_cbf_header(cbf_header):
        header = {
            "fast": 0,
            "slow": 0,
            "length": 0,
            "byte_offset": False,
            "no_compression": False,
        }
        for record in cbf_header.split("\n"):
            if record.startswith("X-Binary-Size-Fastest-Dimension:"):
                header["fast"] = int(record.split()[-1])
            elif record.startswith("X-Binary-Size-Second-Dimension:"):
                header["slow"] = int(record.split()[-1])
            elif record.startswith("X-Binary-Number-of-Elements:"):
                header["length"] = int(record.split()[-1])
            elif record.startswith("X-Binary-Size:"):
                header["size"] = int(record.split()[-1])
            elif "conversions" in record:
                if "x-CBF_BYTE_OFFSET" in record:
                    header["byte_offset"] = True
                elif "x-CBF_NONE" in record:
                    header["no_compression"] = True

        assert header["length"] == header["fast"] * header["slow"]

        return header

    def _start(self):
        """Open the image file, read the image header, copy it into memory
        for future inspection."""

        self._cif_header = FormatCBF.get_cbf_header(self._image_file)

        self._mime_header = ""

        in_binary_format_section = False

        with FormatCBF.open_file(self._image_file, "rb") as fh:
            for record in fh:
                record = record.decode()
                if "--CIF-BINARY-FORMAT-SECTION--" in record:
                    in_binary_format_section = True
                elif in_binary_format_section and record[0] == "X":
                    self._mime_header += record
                if in_binary_format_section and len(record.strip()) == 0:
                    # http://sourceforge.net/apps/trac/cbflib/wiki/ARRAY_DATA%20Category
                    #    In an imgCIF file, the encoded binary data begins after
                    #    the empty line terminating the header.
                    break


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBF.understand(arg))



"""
An implementation of the FormatCBFMiniPilatus image reader for the P6M
detector at PETRA III beamline P14, which has a vertical goniometer.
"""

from __future__ import annotations

import sys

from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus


class FormatCBFMiniPilatusDESY6MSN115(FormatCBFMiniPilatus):
    """A class for reading mini CBF format Pilatus images for 6M SN 115 @ DESY,
    which has a vertical goniometer axis."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if "Detector: PILATUS 6M-F, S/N 60-0115-F" in record:
                return True

        return False

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header, though for miniCBF
        there are limited options for this."""

        return self._goniometer_factory.known_axis((0, 1, 0))


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatusDESY6MSN115.understand(arg))



"""Format object for XDS files"""

from __future__ import annotations

import sys

from rstbx.cftbx.coordinate_frame_converter import coordinate_frame_converter
from rstbx.cftbx.coordinate_frame_helpers import is_recognized_file
from scitbx import matrix

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format
from dxtbx.model import Detector


class FormatXDS(Format):
    """An image reading class for XDS files"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""
        return is_recognized_file(image_file)

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        super().__init__(image_file, **kwargs)

    def _start(self):
        """Open the image file as a cbf file handle, and keep this somewhere
        safe."""

        # Convert the parameters to cbf conventions
        self._convert_to_cbf_convention(self._image_file)

    def _convert_to_cbf_convention(self, xparm_filename):
        """Get the parameters from the XPARM file and convert them to CBF
        conventions.

        Params:
            xparm_handle The handle to the xparm file.

        """
        # Create a coordinate frame converter and extract other quantities
        cfc = coordinate_frame_converter(xparm_filename)
        self._detector_origin = cfc.get("detector_origin")
        self._rotation_axis = cfc.get("rotation_axis")
        self._fast_axis = cfc.get("detector_fast")
        self._slow_axis = cfc.get("detector_slow")
        self._wavelength = cfc.get("wavelength")
        self._image_size = cfc.get("detector_size_fast_slow")
        self._pixel_size = cfc.get("detector_pixel_size_fast_slow")
        self._starting_angle = cfc.get("starting_angle")
        self._oscillation_range = cfc.get("oscillation_range")
        self._starting_frame = cfc.get("starting_frame")
        self._data_range = cfc.get("data_range")
        self._divergence = 0.0
        self._sigma_divergence = cfc.get("sigma_divergence")
        sample_vector = cfc.get("sample_to_source")
        self._beam_vector = tuple(matrix.col(sample_vector))
        self._panel_offset = cfc.get("panel_offset")
        self._panel_size = cfc.get("panel_size")
        self._panel_origin = cfc.get("panel_origin")
        self._panel_fast = cfc.get("panel_fast")
        self._panel_slow = cfc.get("panel_slow")

    def _goniometer(self):
        """Return a working goniometer instance."""
        return self._goniometer_factory.known_axis(self._rotation_axis)

    def _detector(self):
        """Return a working detector instance."""
        if self._panel_origin is not None:
            detector = Detector()
            root = detector.hierarchy()
            root.set_frame(self._fast_axis, self._slow_axis, self._detector_origin)

            i_panel = 0
            for p_offset, p_size, origin, fast, slow in zip(
                self._panel_offset,
                self._panel_size,
                self._panel_origin,
                self._panel_fast,
                self._panel_slow,
            ):
                # ensure mutual orthogonality in presence of numerical rounding errors
                normal = fast.cross(slow)
                slow = normal.cross(fast)
                p = root.add_panel()
                p.set_type("unknown")
                p.set_raw_image_offset(p_offset)
                p.set_image_size(p_size)
                p.set_name("Panel%d" % i_panel)
                p.set_pixel_size(self._pixel_size)
                p.set_frame(fast.elems, slow.elems, origin.elems)
                i_panel += 1
            return detector

        return self._detector_factory.complex(
            self._detector_factory.sensor("unknown"),
            self._detector_origin,
            self._fast_axis,
            self._slow_axis,
            self._pixel_size,
            self._image_size,
            (0, 1.0e9),
        )

    def _beam(self):
        """Return a working beam instance."""
        return self._beam_factory.make_beam(
            sample_to_source=self._beam_vector,
            wavelength=self._wavelength,
            divergence=self._divergence,
            sigma_divergence=self._sigma_divergence,
        )

    def _scan(self):
        """Return a working scan instance."""
        # Set the scan parameters
        if self._data_range is not None:
            image_range = tuple(self._data_range)
        else:
            image_range = (self._starting_frame, self._starting_frame)
        oscillation = (self._starting_angle, self._oscillation_range)

        # Create the scan object
        return self._scan_factory.make_scan(
            image_range,
            0.0,
            oscillation,
            [0] * (image_range[-1] - image_range[0] + 1),
            deg=True,
        )

    def get_raw_data(self):
        """Get the raw image data. For GXPARM.XDS file raise am exception."""
        raise OSError("GXPARM.XDS does not support image data!")


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatXDS.understand(arg))



from __future__ import annotations

import sys

import h5py

from dxtbx.format.FormatHDF5EigerNearlyNexus import FormatHDF5EigerNearlyNexus


class FormatHDF5EigerNearlyNexusSPring8(FormatHDF5EigerNearlyNexus):
    @staticmethod
    def understand(image_file):
        is_nexus = FormatHDF5EigerNearlyNexus.understand(image_file)
        if not is_nexus:
            return False

        # Get the file handle
        with h5py.File(image_file, "r") as handle:
            if "/entry/instrument/detector/detector_number" in handle and handle[
                "/entry/instrument/detector/detector_number"
            ][()] in [b"E-32-0114", b"E-32-0112", b"E-18-0103"]:
                return True
        return False

    def _start(self):
        super()._start()
        # invert the rotation axis
        rotation_axis = self._goniometer_model.get_rotation_axis_datum()
        self._goniometer_model.set_rotation_axis_datum([-x for x in rotation_axis])


if __name__ == "__main__":
    print(FormatHDF5EigerNearlyNexusSPring8.understand(sys.argv[1]))
    print(FormatHDF5EigerNearlyNexusSPring8(sys.argv[1]).get_goniometer())



from __future__ import annotations

from typing import List, Tuple

import h5py
import numpy as np

import cctbx.array_family.flex as flex

from dxtbx import IncorrectFormatError
from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.model import Detector, Goniometer, Scan
from dxtbx.model.beam import BeamFactory, PolychromaticBeam, Probe
from dxtbx.model.goniometer import GoniometerFactory
from dxtbx.model.scan import ScanFactory


class FormatISISSXD(FormatHDF5):
    """
    Class to read image files from ISIS SXD
    (https://www.isis.stfc.ac.uk/Pages/sxd.aspx)
    """

    def __init__(self, image_file: str, **kwargs) -> None:
        super().__init__(image_file, **kwargs)
        if not FormatISISSXD.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        self._nxs_file = h5py.File(image_file, "r")
        self._raw_data = None

    @staticmethod
    def understand(image_file: str) -> bool:
        try:
            return FormatISISSXD.is_isissxd_file(image_file)
        except (OSError, KeyError):
            return False

    @staticmethod
    def is_isissxd_file(image_file: str) -> bool:
        def get_name(image_file: str) -> str:
            with h5py.File(image_file, "r") as handle:
                if "raw_data_1" in handle:
                    return handle["/raw_data_1/name"][0].decode()
                return ""

        return get_name(image_file) == "SXD"

    def get_instrument_name(self) -> str:
        return "SXD"

    def get_experiment_title(self) -> str:
        return self._nxs_file["raw_data_1"]["title"][0].decode()

    def get_experiment_run_number(self) -> str:
        return self._nxs_file["raw_data_1"]["run_number"][0]

    def get_experiment_description(self) -> str:
        title = self.get_experiment_title()
        run_number = self.get_experiment_run_number()
        return f"{title} ({run_number})"

    def get_goniometer_phi_angle(self) -> float:
        try:
            experiment_title = self.get_experiment_title()
            if "w=" in experiment_title:
                return float(experiment_title.split("w=")[1].split()[0])
            elif "wccr" in experiment_title:
                return float(experiment_title.split("wccr=")[1].split()[0])
            elif "wtl" in experiment_title:
                return float(experiment_title.split("wtl=")[1].split()[0])
            else:
                return 0
        except (ValueError, IndexError):
            return 0

    def get_goniometer(self, idx: int = None) -> Goniometer:
        rotation_axis = (0.0, 1.0, 0.0)
        fixed_rotation = (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
        goniometer = GoniometerFactory.make_goniometer(rotation_axis, fixed_rotation)
        angle = self.get_goniometer_phi_angle()
        goniometer.rotate_around_origin(rotation_axis, angle)
        return goniometer

    def get_detector(self, index: int = None) -> Detector:
        num_panels = self._get_num_panels()
        panel_names = self._get_panel_names()
        panel_type = self._get_panel_type()
        image_size = self._get_image_size()
        trusted_range = self._get_panel_trusted_range()
        pixel_size = self._get_pixel_size()
        fast_axes = self._get_panel_fast_axes()
        slow_axes = self._get_panel_slow_axes()
        panel_origins = self._get_panel_origins()
        gain = self._get_panel_gain()
        panel_projections = self._get_panel_projections_2d()
        detector = Detector()
        root = detector.hierarchy()

        for i in range(num_panels):
            group = root.add_group()
            panel = group.add_panel()
            panel.set_type(panel_type)
            panel.set_name(panel_names[i])
            panel.set_image_size(image_size)
            panel.set_trusted_range(trusted_range)
            panel.set_pixel_size(pixel_size)
            panel.set_local_frame(fast_axes[i], slow_axes[i], panel_origins[i])
            panel.set_gain(gain)
            r, t = panel_projections[i + 1]
            r = tuple(map(int, r))
            t = tuple(map(int, t))
            panel.set_projection_2d(r, t)

        return detector

    def _get_pixel_size(self) -> tuple[float, float]:
        # (mm)
        return (3.0, 3.0)

    def _get_image_size(self) -> tuple[int, int]:
        # (px)
        return (64, 64)

    def _get_num_panels(self) -> int:
        return 11

    def _get_panel_names(self) -> list[str]:
        return ["%02d" % (i + 1) for i in range(11)]

    def _get_panel_gain(self):
        return 1.0

    def _get_panel_trusted_range(self) -> tuple[int, int]:
        return (-1, 100000)

    def _get_panel_type(self) -> str:
        return "SENSOR_PAD"

    def _get_panel_origins(self) -> tuple[tuple[float, float, float]]:
        # (mm)

        start_date = self.get_start_date()
        assert start_date is not None
        year, _, _ = start_date.split("-")
        year = int(year)
        if year < 2020:
            return (
                (213.066, -96.0, -120.018),
                (224.999, -96.0, 96.0),
                (60.809, -96.0, 236.945),
                (-214.172, -96.0, 118.198),
                (-224.999, -96.0, -96.0),
                (-60.809, -96.0, -236.945),
                (127.534, -256.614, 96.0),
                (-96.0, -256.614, 127.534),
                (-123.036, -258.801, -96.0),
                (96.0, -256.614, -127.534),
                (96.0, -278.0, 96.0),
            )

        elif year < 2024:
            return (
                (60.81, 96.0, -236.946),
                (224.999, -96.0, 96.0),
                (60.809, -96.0, 236.945),
                (-214.172, -96.0, 118.198),
                (-224.999, -96.0, -96.0),
                (-60.809, -96.0, -236.945),
                (127.534, -256.614, 96.0),
                (-96.0, -256.614, 127.534),
                (-123.036, -258.801, -96.0),
                (96.0, -256.614, -127.534),
                (96.0, -278.0, 96.0),
            )
        else:
            return (
                (60.81, 96.0, -236.946),
                (224.999, 96.0, -96.0),
                (213.065, 96.0, 120.017),
                (-62.876000000000005, 96.0, 236.46999999999997),
                (-224.999, 96.0, 96.0),
                (-213.065, 96.0, -120.017),
                (258.86499037, -123.10256614, -96.0),
                (96.0, -123.10256614147632, 258.86499037410107),
                (-258.78, -123.05699999999999, 96.0),
                (-99.0, -123.10256614147632, -258.86499037410107),
                (96.0, -278.0, 96.0),
            )

    def _panel_0_flipped(self) -> bool:
        start_date = self.get_start_date()
        assert start_date is not None
        year, _, _ = start_date.split("-")
        if int(year) < 2024:
            return True
        return False

    def _get_panel_slow_axes(self) -> tuple[tuple[float, float, float]]:
        start_date = self.get_start_date()
        assert start_date is not None
        year, _, _ = start_date.split("-")
        year = int(year)
        if year < 2020:
            return (
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.695, 0.719, -0.0),
                (0.0, 0.719, 0.695),
                (-0.707, 0.707, -0.0),
                (0.0, 0.719, -0.695),
                (-0.0, 0.0, -1.0),
            )
        elif year < 2024:
            return (
                (0.793, 0.0, 0.609),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.695, 0.719, -0.0),
                (0.0, 0.719, 0.695),
                (-0.707, 0.707, -0.0),
                (0.0, 0.719, -0.695),
                (-0.0, 0.0, -1.0),
            )

        else:
            return (
                (0.0, -1.0, 0.0),
                (0.0, -1.0, 0.0),
                (0.0, -1.0, 0.0),
                (0.0, -1.0, 0.0),
                (0.0, -1.0, 0.0),
                (0.0, -1.0, 0.0),
                (-0.70744243, -0.70676107, 0.0),
                (0.0, -0.7067610703435333, -0.7074424283620987),
                (0.707, -0.707, 0.0),
                (0.0, -0.7067610703435333, 0.7074424283620987),
                (-0.0, 0.0, -1.0),
            )

    def _get_panel_fast_axes(self) -> tuple[tuple[float, float, float]]:
        start_date = self.get_start_date()
        assert start_date is not None
        year, _, _ = start_date.split("-")
        year = int(year)
        if year < 2020:
            return (
                (-0.793, 0.0, -0.609),
                (-0.0, -0.0, -1.0),
                (0.793, -0.0, -0.609),
                (0.788, -0.0, 0.616),
                (-0.0, -0.0, 1.0),
                (-0.793, -0.0, 0.609),
                (0.0, -0.0, -1.0),
                (1.0, -0.0, -0.0),
                (-0.0, -0.0, 1.0),
                (-1.0, -0.0, -0.0),
                (-1.0, -0.0, -0.0),
            )

        elif year < 2024:
            return (
                (0.0, -1.0, 0.0),
                (-0.0, -0.0, -1.0),
                (0.793, -0.0, -0.609),
                (0.788, -0.0, 0.616),
                (-0.0, -0.0, 1.0),
                (-0.793, -0.0, 0.609),
                (0.0, -0.0, -1.0),
                (1.0, -0.0, -0.0),
                (-0.0, -0.0, 1.0),
                (-1.0, -0.0, -0.0),
                (-1.0, -0.0, -0.0),
            )

        else:
            return (
                (0.793, 0.0, 0.609),
                (-0.0, -0.0, 1.0),
                (-0.793, 0.0, 0.609),
                (-0.788, 0.0, -0.616),
                (0.0, 0.0, -1.0),
                (0.793, 0.0, -0.609),
                (0.0, 0.0, 1.0),
                (-1.0, 0.0, 0.0),
                (0.0, 0.0, -1.0),
                (1.0, 0.0, 0.0),
                (-1.0, 0.0, 0.0),
            )

    def _get_panel_projections_2d(self) -> dict:
        """
        Returns a projection of the
        detector flattened around the bottom panel (11)
        """

        p_w, p_h = self._get_image_size()
        panel_pos = {
            11: ((1, 0, 0, 1), (0, 0)),
            10: ((1, 0, 0, 1), (-p_h, 0)),
            8: ((1, 0, 0, 1), (p_h, 0)),
            7: ((1, 0, 0, 1), (0, -p_w)),
            9: ((1, 0, 0, 1), (0, p_w)),
            2: ((1, 0, 0, 1), (0, 2 * -p_w)),
            5: ((1, 0, 0, 1), (0, 2 * p_w)),
            3: ((1, 0, 0, 1), (p_h, 2 * -p_w)),
            4: ((1, 0, 0, 1), (p_h, 2 * p_w)),
            1: ((1, 0, 0, 1), (-p_h, 2 * -p_w)),
            6: ((1, 0, 0, 1), (-p_h, 2 * p_w)),
        }

        return panel_pos

    def get_beam(self, index: int = None) -> PolychromaticBeam:
        direction = self._get_sample_to_source_direction()
        distance = self._get_sample_to_source_distance()
        wavelength_range = self._get_wavelength_range()
        return BeamFactory.make_polychromatic_beam(
            direction=direction,
            sample_to_source_distance=distance,
            probe=Probe.neutron,
            wavelength_range=wavelength_range,
        )

    def _get_sample_to_source_distance(self) -> float:
        # (mm)
        return 8300.0

    def _get_sample_to_source_direction(self) -> tuple[float, float, float]:
        return (0, 0, -1)

    def _get_wavelength_range(self) -> tuple[float, float]:
        # (A)
        return (0.2, 10.0)

    def get_scan(self, index=None) -> Scan:
        image_range = (1, self.get_num_images())
        properties = {
            "time_of_flight": self._get_time_of_flight(),
            "time_of_flight_bin_widths": self._get_time_channel_bin_widths(),
        }
        return ScanFactory.make_scan_from_properties(
            image_range=image_range, properties=properties
        )

    def _get_time_channel_bins(self) -> list[float]:
        # (usec)
        return self._nxs_file["raw_data_1"]["instrument"]["dae"]["time_channels_1"][
            "time_of_flight"
        ][:]

    def _get_time_channel_bin_widths(self) -> List[float]:
        bins = self._get_time_channel_bins()
        return tuple([float((bins[i + 1] - bins[i])) for i in range(len(bins) - 1)])

    def _get_time_of_flight(self) -> Tuple[float]:
        # (usec)
        bins = self._get_time_channel_bins()
        return tuple(
            [float((bins[i] + bins[i + 1]) * 0.5) for i in range(len(bins) - 1)]
        )

    def get_num_images(self) -> int:
        return len(self._get_time_of_flight())

    def _load_raw_data(self):
        raw_data = []
        panel_size = self._get_image_size()
        total_pixels = panel_size[0] * panel_size[1]
        num_tof_bins = len(self._get_time_of_flight())

        # Panel positions are offset by 4 in raw_data array
        # See p24 of https://www.isis.stfc.ac.uk/Pages/sxd-user-guide6683.pdf
        idx_offset = 4

        for panel_idx in range(self._get_num_panels()):
            start_idx = (panel_idx * total_pixels) + (panel_idx * idx_offset)
            end_idx = start_idx + total_pixels
            panel_data = flex.int(
                self._nxs_file["raw_data_1/detector_1/counts"][0, start_idx:end_idx, :]
            )
            panel_data.reshape(flex.grid(panel_size[0], panel_size[1], num_tof_bins))
            raw_data.append(panel_data)
        self._raw_data = tuple(raw_data)

    def get_raw_data(self, index: int, use_loaded_data=True) -> tuple[flex.int]:
        raw_data = []

        if use_loaded_data:
            if self._raw_data is None:
                self._load_raw_data()
            for panel in self._raw_data:
                data = panel[:, :, index : index + 1]
                data.reshape(flex.grid(panel.all()[0], panel.all()[1]))
                data.matrix_transpose_in_place()
                raw_data.append(data)

        else:
            panel_size = self._get_image_size()
            total_pixels = panel_size[0] * panel_size[1]

            # Panel positions are offset by 4 in raw_data array
            # See p24 of https://www.isis.stfc.ac.uk/Pages/sxd-user-guide6683.pdf
            idx_offset = 4

            for panel_idx in range(self._get_num_panels()):
                start_idx = (panel_idx * total_pixels) + (panel_idx * idx_offset)
                end_idx = start_idx + total_pixels
                panel_data = flex.int(
                    self._nxs_file["raw_data_1/detector_1/counts"][
                        0, start_idx:end_idx, index
                    ]
                )
                panel_data.reshape(flex.grid(panel_size[0], panel_size[1]))
                panel_data.matrix_transpose_in_place()
                raw_data.append(panel_data)

        return tuple(raw_data)

    def get_flattened_data(
        self, image_range: None | tuple = None, scale_data: bool = True
    ) -> tuple[flex.int]:
        """
        Image data summed along the time-of-flight direction
        """

        panel_size = self._get_image_size()
        total_pixels = panel_size[0] * panel_size[1]
        # Panel positions are offset by 4 in raw_data array
        # See p24 of https://www.isis.stfc.ac.uk/Pages/sxd-user-guide6683.pdf
        idx_offset = 4
        max_val = None
        num_tof_bins = len(self._get_time_of_flight())
        raw_data = []
        for panel_idx in range(self._get_num_panels()):
            start_idx = (panel_idx * total_pixels) + (panel_idx * idx_offset)
            end_idx = start_idx + total_pixels
            panel_data = self._nxs_file["raw_data_1/detector_1/counts"][
                0, start_idx:end_idx, :
            ]
            panel_data = np.reshape(
                panel_data, (panel_size[0], panel_size[1], num_tof_bins)
            )
            if image_range is not None:
                assert len(image_range) == 2, (
                    "expected image_range to be only two values"
                )
                assert image_range[0] >= 0 and image_range[0] < image_range[1], (
                    "image_range[0] out of range"
                )
                assert image_range[1] <= num_tof_bins, "image_range[1] out of range"
                panel_data = np.flipud(
                    np.sum(panel_data[:, :, image_range[0] : image_range[1]], axis=2)
                )
            else:
                panel_data = np.flipud(np.sum(panel_data, axis=2))
            if panel_idx == 0 and self._panel_0_flipped():
                panel_data = np.flipud(panel_data)
            panel_max_val = np.max(panel_data)
            if max_val is None or max_val < panel_max_val:
                max_val = panel_max_val
            raw_data.append(panel_data)

        if scale_data:
            return tuple([(i / max_val).tolist() for i in raw_data])

        return tuple([i.tolist() for i in raw_data])

    def get_flattened_pixel_data(
        self, panel_idx: int, x: int, y: int
    ) -> tuple[tuple, tuple]:
        time_channels = self._get_time_of_flight()
        panel_size = self._get_image_size()
        height = panel_size[1]
        total_pixels = panel_size[0] * panel_size[1]
        # Panel positions are offset by 4 in raw_data array
        # See p24 of https://www.isis.stfc.ac.uk/Pages/sxd-user-guide6683.pdf
        idx_offset = 4
        idx = (panel_idx * total_pixels) + (panel_idx * idx_offset) + y * height + x
        return (
            time_channels,
            tuple(self._nxs_file["raw_data_1/detector_1/counts"][0, idx, :].tolist()),
        )

    def get_proton_charge(self) -> float:
        return float(self._nxs_file["raw_data_1/proton_charge"][0])

    def get_start_date(self) -> str:
        """
        Date of experiment in YYYY-MM-DD format
        """
        try:
            raw_date = self._nxs_file["raw_data_1/start_time"][0].decode()
            return raw_date.split("T")[0]
        except IndexError:
            return None



"""Format class to recognise images from a ThermoFisher Ceta D detector that
have been converted to an ADSC-like SMV with useful metadata. We
want to override the beam model to produce an unpolarised beam and to set the
detector gain to something sensible."""

from __future__ import annotations

import os

from dxtbx.format.FormatSMVADSC import FormatSMVADSC
from dxtbx.model.beam import Probe


# We have to inherit from FormatSMVADSC rather than FormatSMV, so that we can
# activate this class with an enivronment variable in cases where the
# BEAMLINE or DETECTOR_SN are not set to anything specific (for example files
# that lie and set BEAMLINE=ALS831).
class FormatSMVADSCCetaD(FormatSMVADSC):
    @staticmethod
    def understand(image_file):
        # Allow this class to override FormatSMVADSC with an environment variable
        if "FORCE_SMV_AS_CETAD" in os.environ:
            return True

        # Otherwise recognise specific instruments from the header
        size, header = FormatSMVADSC.get_smv_header(image_file)
        check = header.get("BEAMLINE", "") + header.get("DETECTOR_SN", "")
        if "CETAD" in check:
            return True

        return False

    def _detector(self):
        distance = float(self._header_dictionary["DISTANCE"])
        beam_x = float(self._header_dictionary["BEAM_CENTER_X"])
        beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )
        # Ceta has gain of > 26 and saturates at about 8000.0 for binning=1
        # according to Thermo Fisher
        binning = {"1x1": 1, "2x2": 2}.get(self._header_dictionary.get("BIN"), 1)
        gain = float(self._header_dictionary.get("GAIN", 26.0))
        saturation = 8000 * binning**2
        trusted_range = (-1000, saturation)
        pedestal = float(self._header_dictionary.get("IMAGE_PEDESTAL", 0))

        return self._detector_factory.simple(
            "PAD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            trusted_range,
            [],
            gain=gain,
            pedestal=pedestal,
        )

    def _beam(self):
        """Return a simple model for an unpolarised beam."""

        wavelength = float(self._header_dictionary["WAVELENGTH"])
        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )
